from ext_auth.backends.providers.aad import get_token
from django.contrib.auth import logout
from django.conf import settings

        
def graph_token_middleware(get_response):
    def middleware(request):
        # Before view is called
        token = get_token(request)
        if not token and settings.EXT_AUTH_AAD_AUTH_FLOW_KEY not in request.session:
            # If no access token is present we need to login again
            logout(request)

        request.graph_token = token

        response = get_response(request)
        # After view is called

        return response

    return middleware



from typing import Union
from msal import (
    SerializableTokenCache,
    ConfidentialClientApplication
)

from django.contrib.auth.models import User
from django.conf import settings

#import ext_auth.services.ms_graph as ms_graph
from ext_auth.services.ms_graph import get_graph_user
from ext_auth.backends.ext_auth import ExtAuthBackend, get_ext_auth_backend, AuthenticationException
from ext_auth.choices import ExternalAuthType
from django.contrib.auth import get_user_model
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)

UserModel = get_user_model()

def load_token_cache(request):
    # Check for a token cache in the session
    cache = SerializableTokenCache()
    if request.session.get(settings.EXT_AUTH_AAD_TOKEN_CACHE_KEY):
        cache.deserialize(
            request.session[settings.EXT_AUTH_AAD_TOKEN_CACHE_KEY])

    return cache

def save_token_cache(request, cache):
    # If cache has changed, persist back to session
    if cache.has_state_changed:
        request.session[settings.EXT_AUTH_AAD_TOKEN_CACHE_KEY] = cache.serialize()

def clear_session(request):
    """
    Clears the current users session data, which will force new auth
    """
    request.session.pop(settings.EXT_AUTH_AAD_TOKEN_CACHE_KEY)

def has_external_auth(request) -> bool:
    return bool(request.session.get('user'))

def validate_user_dict(user_dict: dict) -> bool:
    return bool(user_dict.get('userPrincipalName'))

def get_sign_in_flow(request):
    backend = get_ext_auth_backend(request)
    return backend.get_sign_in_flow(request)

def get_token(request) -> str:
    backend = get_ext_auth_backend(request)

    return backend.get_token(request)


class AzureADBackend(ExtAuthBackend):
    """
    Implementation of Azure AD single tenant authentication
    """

    ext_auth_type = ExternalAuthType.AZURE_AD

    def get_redirect_uri(self, request) -> str:
         # Get the sign-in flow
        flow = get_sign_in_flow(request)
        # Save the expected flow, so we can use it in the callback
        try:
            request.session[settings.EXT_AUTH_AAD_AUTH_FLOW_KEY] = flow
        except Exception as e:
            print(e)
        # Redirect to the Azure sign-in page
        if 'next' in request.GET:
            request.session['ext_auth_next'] = request.GET['next']

        return flow['auth_uri']

    def ext_authenticate(self, request, **kwargs) -> dict:
        flow = request.session.pop(settings.EXT_AUTH_AAD_AUTH_FLOW_KEY, {})
        if 'code' not in request.GET:
            raise AuthenticationException("Tried to authenticate, but auth code was not in request")
        try:
            result = self.get_token_from_code(request, flow)
            if settings.EXT_AUTH_AAD_ACCESS_TOKEN_KEY in result:
                token = result.get(
                        settings.EXT_AUTH_AAD_ACCESS_TOKEN_KEY)
                return self.get_ext_user(request, token)
            elif 'error' in result:
                error_msg = f"""
                        Failed to authenticate using msal app: {result.get('error')} \n
                        Dict error result {result.__str__()}
                """
                raise AuthenticationException(error_msg)
            else:
                raise AuthenticationException("""No valid result, not error received,
                authentication not succesfull for unknown reason""")
        except ValueError as e:
            logger.error("Could not get token from code...")
            logger.error(e)
            raise AuthenticationException("Got value error while getting result from msal app...")

    def get_ext_user(self, request, token, **kwargs) -> dict:
        graph_user = get_graph_user(token)
        return {
            'username': graph_user.get('userPrincipalName'),
            'email': graph_user.get('userPrincipalName'),
            'firstName': graph_user.get('givenName'),
            'lastName': graph_user.get('surname'),
            'department': graph_user.get('department')
        }

    def client_id(self, request):
        return settings.EXT_AUTH_AAD_CLIENT_ID

    def client_secret(self, request):
        return settings.EXT_AUTH_AAD_CLIENT_SECRET

    def authority(self, request):
        return settings.EXT_AUTH_AAD_AUTH_AUTHORITY

    def get_msal_app(self, cache=None, request=None) -> ConfidentialClientApplication:
        # Initialize the MSAL confidential client
        auth_app = ConfidentialClientApplication(
            client_id=self.client_id(request),
            authority=self.authority(request),
            client_credential=self.client_secret(request),
            token_cache=cache)

        return auth_app

    def get_sign_in_flow(self, request):
        auth_app = self.get_msal_app(request=request)

        return auth_app.initiate_auth_code_flow(
            settings.EXT_AUTH_AAD_SCOPES,
            redirect_uri=settings.EXT_AUTH_AAD_REDIRECT_URI)

    # Method to exchange auth code for access token
    def get_token_from_code(self, request, flow):
        cache = load_token_cache(request)
        auth_app = self.get_msal_app(cache)
        # Get the flow saved in session
        result = auth_app.acquire_token_by_auth_code_flow(
            flow, request.GET)

        save_token_cache(request, cache)
        return result

    def get_token(self, request) -> Union[str, None]:
        cache = load_token_cache(request)
        auth_app = self.get_msal_app(cache)

        accounts = auth_app.get_accounts()
        if accounts:
            result = auth_app.acquire_token_silent(
                settings.EXT_AUTH_AAD_SCOPES,
                account=accounts[0])

            # This can happen for various reasons, error or token not present in cache
            if not result:
                logger.error("No result from acquire_token_silent...")
                auth_app.remove_account(accounts[0])
                clear_session(request)
                return

            # If for some weird reason, token is not in the validated dict
            token = result.get(settings.EXT_AUTH_AAD_ACCESS_TOKEN_KEY)
            if not token:
                logger.error("No token from acquire_token_silent...")
                auth_app.remove_account(accounts[0])
                clear_session(request)
                return

            # If everything went well, save the new state to session data
            save_token_cache(request, cache)
            return token



from abc import ABC, abstractmethod
import secrets
import string

from django.contrib.auth import get_user_model, _get_backends
from django.http import HttpResponseRedirect
from django.contrib.auth.backends import ModelBackend

from ext_auth.models import UserProfile
from ext_auth.choices import ExternalAuthType

UserModel = get_user_model()

def make_random_password(length=16):
    alphabet = string.ascii_letters + string.digits + string.punctuation
    password = ''.join(secrets.choice(alphabet) for _ in range(length))
    return password


class AuthenticationException(Exception):
    pass


class ExtAuthBackend(ABC, ModelBackend):
    """
    Abstract Class that defines a set of methods that needs
    to be implemented for evert external authentication
    provider.
    """

    ext_auth_type: ExternalAuthType

    def create_user(self, username: str, email: str, **kwargs) -> UserModel:
        user = UserModel.objects.create_user(
            username=username,
            email=email,
            password=make_random_password(length=14)
        )
        user_profile = UserProfile(
            user=user,
            display_name=" ".join([user.first_name, user.last_name]),
            department="",
            ext_auth=self.ext_auth_type
        )
        self.save_user_profile(user_profile)
        return user
        

    def save_user_profile(self, user_profile: UserProfile) -> UserProfile:
        user_profile = user_profile.save()
        return user_profile

    def user_exists(self, email: str, **kwargs) -> bool:
        try:
            UserModel.objects.get(email=email)
            return True
        except UserModel.DoesNotExist:
            return False

    def init_auth(self, request) -> HttpResponseRedirect:
        redirect = self.get_redirect_uri(request)
        return HttpResponseRedirect(redirect)

    def authenticate(self, request, username=None, password=None, **kwargs):
        """
        Authenticates the user using ext auth backend
        Can raise a AuthenticationException
        """
        user_dict = self.ext_authenticate(request, **kwargs)
        if not user_dict or 'email' not in user_dict:
            raise AuthenticationException("Authentication did not return a valid user dict...")
        # user_dict should have username and email keys
        if not self.user_exists(user_dict.get('email')):
            user = self.create_user(
                user_dict.get('username'),
                user_dict.get('email')
            )
            return user

        return self.get_user_by_username(user_dict.get('username'))

    def get_user_by_username(self, username):
        try:
            user = UserModel._default_manager.get(username=username)
        except UserModel.DoesNotExist:
            return None
        return user if self.user_can_authenticate(user) else None

    @abstractmethod
    def get_redirect_uri(request) -> str:
        ...

    @abstractmethod
    def ext_authenticate(self, request, **kwargs) -> dict:
        """
        This method should handle authenticating the user.
        It is being called from the ModelBackend's own
        authenticate method, which again gets called on a redirect
        callback.

        Should just return dict representation of the
        external authentication response:
        key: email
        key: username
        """
        ...

    @abstractmethod
    def get_ext_user(self, request, **kwargs) -> dict:
        """
        Simply fetches the user from this authentication provider.
        """
        ...


def get_ext_auth_backend(request, type: ExternalAuthType = None) -> ExtAuthBackend:
    for backend, backend_path in _get_backends(return_tuples=True):
        if isinstance(backend, ExtAuthBackend):
            if not type:
                return backend
            
            # If type spe
            if type == backend.ext_auth_type:
                return backend




from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required
from ext_auth.backends.ext_auth import get_ext_auth_backend, AuthenticationException
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


def sign_in(request):
    backend = get_ext_auth_backend(request)
    http_response_redirect = backend.init_auth(request)

    if 'next' in request.GET:
        request.session['ext_auth_next'] = request.GET['next']

    # Redirect to ext auth signin
    return http_response_redirect


@login_required
def sign_out(request):
    logout(request)
    return HttpResponseRedirect(reverse("home"))


def auth_failed_error() -> HttpResponse:
    html_content = b"<html><body><h1>Authentication failed...</h1></body></html>"
    return HttpResponse(html_content)


def callback(request):
    # callback with code received use it to authenticate the user
    try:
        user = authenticate(request)
        if not user:
            logger.error("Did not get a user dict back from auth backend...")
            return auth_failed_error()

        login(request, user)

        if 'ext_auth_next' in request.session:
            return HttpResponseRedirect(request.session.pop('ext_auth_next'))

        return HttpResponseRedirect(settings.EXT_AUTH_POST_LOGIN_REDIRECT_URI)
    except AuthenticationException as e:
        logger.error("Failed to authenticate in callback...")
        logger.error(e)
        return auth_failed_error()




# Generated by Django 4.1 on 2022-12-20 12:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ext_auth', '0002_userprofile_display_name_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='userprofile',
            name='mode',
        ),
    ]



from django.test import TestCase
from django.contrib.auth import get_user_model
from unittest.mock import Mock, patch

from ext_auth.backends import ExtAuthBackend, AzureADBackend
import ext_auth.backends.ext_auth as ext_auth
from ext_auth.choices import ExternalAuthType

UserModel = get_user_model()

class TestBackend(ExtAuthBackend):
    ext_auth_type = -1

    def get_redirect_uri(request) -> str:
        ...

    def ext_authenticate(self, request, **kwargs) -> dict:
        ...

    def get_ext_user(self, request, **kwargs):
        ...

USER_DICT = {
    'userPrincipalName': 'johndoe@example.com',
    'givenName': 'John',
    'surname': 'Doe',
    'displayName': 'John Doe',
    'department': 'Sales'
}

class ExtAuthBackendTests(TestCase):

    @patch('ext_auth.backends.ext_auth._get_backends')
    def test_get_ext_auth_backend(self, mock_get_backends):
        
        request = Mock()
        backend = AzureADBackend()
        mock_get_backends.return_value = [(backend, Mock())]
        
        result = ext_auth.get_ext_auth_backend(request)
        self.assertEqual(result,backend)

    @patch('ext_auth.backends.ext_auth._get_backends')
    def test_get_specific_ext_auth_backend(self, mock_get_backends):
        request = Mock()

        aad_backend = AzureADBackend()
        test_backend = TestBackend()
        mock_get_backends.return_value = [(aad_backend, Mock()), (test_backend, Mock())]
        
        test_result = ext_auth.get_ext_auth_backend(request, -1)
        aad_result = ext_auth.get_ext_auth_backend(request, ExternalAuthType.AZURE_AD)
        none_result = ext_auth.get_ext_auth_backend(request, -2)
        self.assertEqual(aad_result, aad_backend)
        self.assertEqual(test_result, test_backend)
        self.assertEqual(none_result, None)
    
    def test_create_get_user(self):

        username = 'leomessi'
        email = 'leomessi@gmail.com' 

        aad_backend = AzureADBackend()
        result_user = aad_backend.create_user(username, email)
        self.assertEqual(result_user.username, username)
        self.assertEqual(result_user.email, email)
        
        get_user_result = aad_backend.get_user(result_user.id)
        self.assertEqual(get_user_result, result_user)

    def test_get_user_by_email(self):

        username = 'leomessi'
        email = 'leomessi@gmail.com' 
        aad_backend = AzureADBackend()

        aad_backend.create_user(username, email)

        get_user_by_username_result = aad_backend.get_user_by_username(username)
        self.assertEqual(get_user_by_username_result.username, username)
    
    @patch(
        'ext_auth.backends.providers.aad.'
        'AzureADBackend.ext_authenticate'
    )
    def test_authenticate(self, mock_ext_authenticate):
        request = Mock()
        aad_backend = AzureADBackend()
        mock_ext_authenticate.return_value = {
            'username': USER_DICT.get('userPrincipalName'),
            'email': USER_DICT.get('userPrincipalName'),
            'firstName': USER_DICT.get('givenName'),
            'lastName': USER_DICT.get('surname'),
            'department': USER_DICT.get('department')
        }
        with patch(
            'ext_auth.backends.ext_auth.ExtAuthBackend.create_user'
        ) as mock_create_user:
            aad_backend.authenticate(request)
            mock_create_user.assert_called_once()
            mock_create_user.assert_called_with(
                USER_DICT.get('userPrincipalName'),
                USER_DICT.get('userPrincipalName')
            )
    
    def test_create_get_user_profile(self):

        username = 'leomessi'
        email = 'leomessi@gmail.com' 

        aad_backend = AzureADBackend()
        result_user = aad_backend.create_user(username, email)
        self.assertEqual(result_user.userprofile.ext_auth, ExternalAuthType.AZURE_AD)

    def test_user_exists(self):
        username = 'leomessi'
        email = 'leomessi@gmail.com' 

        aad_backend = AzureADBackend()
        aad_backend.create_user(username, email)
        user_exists = aad_backend.user_exists(email)
        # Test that user exists
        self.assertEqual(user_exists, True)
        # Test that it does not exist
        user_does_not_exist = aad_backend.user_exists('shouldnotexist')
        self.assertEqual(user_does_not_exist, False)

    def test_random_password(self):
        pw = ext_auth.make_random_password()
        self.assertEqual(len(pw), 16)



# Generated by Django 4.1 on 2022-10-02 20:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ext_auth', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='userprofile',
            name='display_name',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='userprofile',
            name='department',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



import requests
from itertools import chain
from django.conf import settings
graph_url = "https://graph.microsoft.com/v1.0"


def has_advanced_user_info():
    return "user.read.all" in settings.EXT_AUTH_AAD_SCOPES


def get_user_query_params():
    if has_advanced_user_info():
        return "displayName,userPrincipalName,profilePhoto,department,givenName,surname"
    return "displayName,userPrincipalName"


def get_graph_user(token) -> dict:
    # Send GET to /me
    user = requests.get(
        "{0}/me".format(graph_url),
        headers={"Authorization": "Bearer {0}".format(token)},
        params={
            "$select": get_user_query_params()},
    )
    # Return the JSON result
    json_user = user.json()
    return json_user


def get_all_users(token):
    users = []
    next_link = f"{graph_url}/users"
    params = {"$select": get_user_query_params()}
    while next_link:
        new_users = requests.get(
            next_link,
            headers={"Authorization": f"Bearer {token}"},
            params=params,
        ).json()

        next_link = new_users.get("@odata.nextLink")
        users = chain(users, new_users.get("value", []))
        params = None

    return list(users)


def id_in_group(token, guid):
    response = requests.get(
        "{0}/me/memberOf".format(graph_url),
        headers={"Authorization": "Bearer {0}".format(token)},
        params={"$filter": "id eq '" + guid + "'"},
    )

    if "value" in response.json():
        for value in response.json()["value"]:
            if "id" in value:
                if value["id"] == guid:
                    return True

    return False


def get_profile_pic(token):
    profile_picture = requests.get(
        "{0}//me/photo/$value".format(graph_url),
        headers={"Authorization": "Bearer {0}".format(token)},
        params={},
    )
    return profile_picture


def get_graph_user_by_email(token, email) -> dict:
    # Send GET to /me
    user = requests.get(
        "{0}/users('{1}')".format(graph_url, email),
        headers={"Authorization": "Bearer {0}".format(token)},
        params={
            "$select": get_user_query_params()},
    )
    # Return the JSON result
    json_user = user.json()
    return json_user


def user_in_group(token, guid):
    if guid is None:
        return False

    response = requests.get(
        '{0}/me/memberOf'.format(graph_url),
        headers={
            'Authorization': 'Bearer {0}'.format(token)
        },
        params={
            '$filter': "id eq '" + guid + "'"
        })

    for value in response.json().get('value', []):
        if value.get('id', None) == guid:
            return True

    return False
